CAMPUS_KEYPOINTS = [
    'right_ankle', 'right_knee', 'right_hip_extra', 'left_hip_extra',
    'left_knee', 'left_ankle', 'right_wrist', 'right_elbow', 'right_shoulder',
    'left_shoulder', 'left_elbow', 'left_wrist', 'jaw', 'headtop'
]
